/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.
****************************************************************************************/
// OLED pins to ESP32 GPIOs via this connection:
// OLED_SDA -- GPIO5
// OLED_SCL -- GPIO4
// OLED_RST -- GPIO16

// DS18B20 - GPIO25 - Data

// SD - GPIO12 - SCL
// SD - GPIO15 - MISO 
// SD - GPIO13 - MOSI
// SD - GPIO14 - CS

#include "OneWire.h"
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "display1306.h"
#include "sensor18b20.h"
#include "handleSD.h"
#include "tempLoggerESP32.h"
#include "myNTPClient.h"
#include "WebDebug.h"

//////////////////////////////////////////////////////////////////////
SemaphoreHandle_t  m_xBasecampSemaphore;
bool connection = false;

TaskHandle_t OledTask = NULL;
TaskHandle_t TempTask = NULL;
extern TaskHandle_t _async_service_task_handle;

//////////////////////////////////////////////////////////////////////
Basecamp& MyBasecamp::GetInstance()
{
  static Basecamp instance;
  return instance;
}

//////////////////////////////////////////////////////////////////////
///void main_task(void * pvParameters)
void setup()
{
  m_xBasecampSemaphore = xSemaphoreCreateMutex();
  
  Basecamp& iot = MyBasecamp::GetInstance();
  iot.begin();
  Serial.begin(115200);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }

  Serial.printf("Temperature logger, V%s\n", VERSION_STRING);

  display1306& display = display1306::GetInstance();
 
  display.addMessage(String("Initialise... " VERSION_STRING "\n"), PRIO_INFO);

  // start sensor detection 
  xTaskCreatePinnedToCore(
    sensor18b20::RunInstance,       // Function to implement the task 
    "tempTask",                     // Name of the task 
    4096,                           // Stack size in words  
    NULL,                           // Task input parameter 
    7,                              // Priority of the task 
    &TempTask,                      // Task handle. 
    1);                             // Core where the task should run 

  // wait for finished detection
  while(!sensor18b20::GetInstance().isSetupFinished())
  {
    delay (100);
  }
 
  // start oled task
  xTaskCreatePinnedToCore(
    display1306::RunInstance,     // Function to implement the task 
    "uiTask",                     // Name of the task 
    4096*2,                         // Stack size in words
    NULL,                         // Task input parameter 
    5,                            // Priority of the task 
    &OledTask,                     // Task handle. 
    1);                           // Core where the task should run 
    
  if (iot.configuration.get("WifiConfigured") != "True") 
  {
    display.addMessage(String("Wifi not configured!\nConnect to hotspot 'ESP32'\nand open 192.168.4.1\n Version: " VERSION_STRING), PRIO_ERROR);
  } 
  else
  {
    int retry = 0;
    while ((WiFi.status() != WL_CONNECTED) && (retry < 30)) 
    {
      Serial.printf("### WIFI: retry=%d\n", retry);
      retry++;
      delay(500);
    }
    if (retry == 30 )
    {
      connection = false;
      display.addMessage(String("Could not connect to\n") + iot.configuration.get("WifiEssid"), PRIO_ERROR);
    } 
    else 
    {
      connection = true;
      IPAddress ip = WiFi.localIP(); 
      String strIP = String(ip[0]) + "." + String(ip[1]) + "." + String(ip[2]) + "." + String(ip[3]) + " - " + iot.configuration.get("DeviceName");
      display.addMessage(String("Connect to ") + iot.configuration.get("WifiEssid") + String("\n") + String(strIP) + String("\nVersion: " VERSION_STRING), PRIO_INFO);

      // first call of get instance, additional initialise
      MyNTPClient::GetInstance();
    }
  }

  handleSD::GetInstance().setup();

  // add own functions to webserver
  iot.web.addURL("", "", "");
  
  //delay(3000); // auf die ersten Sensor Werte warten
  display.initDone();
}

//////////////////////////////////////////////////////////////////////
void loop()
{
  MyNTPClient::GetInstance().update();
  
  // copy sensor value every 2 seconds into strings 
  sensor18b20::GetInstance().update();

  handleSD::GetInstance().writeToSD();

  delay(2000);
}

//////////////////////////////////////////////////////////////////////
