/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.
****************************************************************************************/
   
#include <SPIFFS.h>
#include <ESPAsyncWebServer.h>
#include "tempLoggerESP32.h"
#include "myNTPClient.h"
#include "display1306.h"
#include "sensor18b20.h"
#include "handleSD.h"
#include "webdebug.h"

#include <arduino.h>
#include "FS.h"
#include "SD.h"
#include "SPI.h"

// read/write data.json
SemaphoreHandle_t m_SemaphoreDataJson;
File g_fileDataJson = (fs::File)0;
size_t g_fSizeDataJson = 0;

SemaphoreHandle_t m_SemaphoreFile;
File g_fileTempData = (fs::File)0;
size_t g_fSizeTempData = 0;

//////////////////////////////////////////////////////////////////////
// callback function 
static size_t readDataJsonFromSD(uint8_t *buffer, size_t maxLen, size_t alreadySent)
{
  if(!g_fileDataJson)
  {
    return 0;
  }
  
  size_t fSize = g_fileDataJson.size();
  if ((fSize-alreadySent)>maxLen) 
  {
    //Serial.printf("read %d\n", maxLen);
    // We have more to read than fits in maxLen Buffer
    g_fileDataJson.read(buffer, maxLen);
    return maxLen;
  }
  // Ok, last chunk
  //Serial.printf("read %d\n", fSize-alreadySent);
  g_fileDataJson.read(buffer, maxLen);
  g_fileDataJson.close();
  xSemaphoreGive(m_SemaphoreDataJson); // release semaphore
  return fSize-alreadySent; // Return from here to end of indexhtml
}

//////////////////////////////////////////////////////////////////////
// callback function 
static size_t readTempDataFromSD(uint8_t *buffer, size_t maxLen, size_t alreadySent)
{
  if(!g_fileTempData)
  {
    return 0;
  }
  
  size_t fSize = g_fileTempData.size();
  if ((fSize-alreadySent)>maxLen) 
  {
    //Serial.printf("read %d\n", maxLen);
    // We have more to read than fits in maxLen Buffer
    g_fileTempData.read(buffer, maxLen);
    return maxLen;
  }
  // Ok, last chunk
  //Serial.printf("read %d\n", fSize-alreadySent);
  g_fileTempData.read(buffer, maxLen);
  g_fileTempData.close();

  handleSD::GetInstance().releaseSaveFileAccess(); // release semaphore read/write same file
  xSemaphoreGive(m_SemaphoreFile); // release semaphore web request
  return fSize-alreadySent; // Return from here to end of indexhtml
}

//////////////////////////////////////////////////////////////////////
bool WebServer::addURL(const char* url, const char* content, const char* mimetype)
{
  Basecamp& iot = MyBasecamp::GetInstance();
  Configuration &configuration = iot.configuration;

  ////////////////////////////////////////////////////////////////////////
  // create semaphores
   m_SemaphoreDataJson = xSemaphoreCreateMutex();
   m_SemaphoreFile = xSemaphoreCreateMutex();
  
  // write data.json
  g_fileDataJson = SD.open("/data.json", FILE_WRITE);

  g_fileDataJson.printf("{\n\"elements\": [\n");
  int elementIndex = 0;
  for (const auto &interfaceElement : interfaceElements)
  {
    std::map<String, String, cmp_str> attributeMap;

    g_fileDataJson.printf("{\n");
    g_fileDataJson.printf("\"attributes\": {\n");
    for (const auto &attribute : interfaceElement.attributes)
    {
      String first = String{attribute.first};
      String second = String{attribute.second};
      attributeMap[first] = second;
    }
     
    if (interfaceElement.getAttribute("data-config").length() != 0)
    {
      if (interfaceElement.getAttribute("type")=="password")
      {
        attributeMap["placeholder"] = "Password unchanged";
        attributeMap["value"] = "";
      } 
      else 
      {
        attributeMap["value"] = String{configuration.get(interfaceElement.getAttribute("data-config"))};
      }
    }
    
    int attributeIndex = 0;
    for (const auto &attribute : attributeMap)
    {
      String first = String{attribute.first};
      String second = String{attribute.second};
      char cSep = ' ';

      // letztes Element
      if(++attributeIndex < attributeMap.size())
      {
        cSep = ',';
      }
   
      g_fileDataJson.printf("\"%s\": \"%s\"%c\n", &first[0], &second[0], cSep);
    }
    g_fileDataJson.printf("},\n");
 
    String item = interfaceElement.element; g_fileDataJson.printf("\"element\": \"%s\",\n", &item[0]);
    item = interfaceElement.id; g_fileDataJson.printf("\"id\": \"%s\",\n", &item[0]);
    item = interfaceElement.content; g_fileDataJson.printf("\"content\": \"%s\",\n", &item[0]);
    item = interfaceElement.parent; g_fileDataJson.printf("\"parent\": \"%s\"\n", &item[0]);

    if(++elementIndex < interfaceElements.size())
    {
      g_fileDataJson.printf("},\n");
    }
    else
    {
      g_fileDataJson.printf("}\n");
    }
  }
  g_fileDataJson.printf("]\n}\n");
  g_fileDataJson.close();

  // write data.json end

  ////////////////////////////////////////////////////////////////////////
  server.on("/data.json" , HTTP_GET, [&configuration, this](AsyncWebServerRequest * request)
  {
    if(xSemaphoreTake(m_SemaphoreDataJson, 0) == pdTRUE )
    {
      g_fileDataJson = SD.open("/data.json", FILE_READ);
      
      if(g_fileDataJson)
      {
        // close file in callback readDataJsonFromSD
        AsyncWebServerResponse *response = request->beginResponse(String("application/json"), g_fileDataJson.size(), readDataJsonFromSD);
        request->send(response);
        return;
      }
      xSemaphoreGive(m_SemaphoreDataJson); // can not open file, release semaphore, return 404
    }
    request->send(404); // if no semaphore return 404
  });

  ////////////////////////////////////////////////////////////////////////
  server.on("/config" , HTTP_GET, [](AsyncWebServerRequest * request)
  {
    Serial.printf("/config\n");
    String stringReturn = sensor18b20::GetInstance().getWebAppConfigText();
    //Serial.printf("app.config %s\n", &stringReturn[0]);
    
    request->send(200, "text/plain", &stringReturn[0]);
  });
  
  ////////////////////////////////////////////////////////////////////////
  server.on("/file" , HTTP_GET, [](AsyncWebServerRequest * request)
  {
    int paramsNr = request->params();
    DateTime dt(MyNTPClient::GetInstance().getEpochTime());
    unsigned long start = millis();
 
    int year = dt.year();
    int month = dt.month();
    int day = dt.day();
    sensor18b20& sensorInstance = sensor18b20::GetInstance();
    String id = "";
    if( sensorInstance.getNumberSensorCnt() > 0)
    {
      id = sensorInstance.getSensorCRC16Id(0);
    }
    
    for(int i=0;i<paramsNr;i++)
    {
      AsyncWebParameter* p = request->getParam(i);
      //Serial.print("Param name: ");
      //Serial.println(p->name());
      //Serial.print("Param value: ");
      //Serial.println(p->value());
      //Serial.println("------");

      if(p->name() == "y")
      {
        year = p->value().toInt();
      }
      else if(p->name() == "m")
      {
        month = p->value().toInt();
      }
      else if(p->name() == "d")
      {
        day = p->value().toInt();
      }
      else if(p->name() == "id")
      {
        id = p->value();
      }
    }
    Serial.printf("/file '%s'\n", &id[0]);

    // protect from other web request's
    if(xSemaphoreTake(m_SemaphoreFile, 2000/portTICK_PERIOD_MS ) == pdTRUE ) // wait max 2 seconds
    {
      char fileName[21];
      snprintf(fileName, sizeof(fileName), "/%d/%02d/%02d_%s.txt", year, month, day, &id[0]);
      Serial.printf("/file '%s'\n", fileName);
 
      // check if file is now open for writing
      if(handleSD::GetInstance().takeSaveFilesAccess(100/portTICK_PERIOD_MS)) // wait max 100 ms for file access
      {
        Serial.printf("/open file '%s'\n", fileName);
        g_fileTempData = SD.open(fileName, FILE_READ);
        if(g_fileTempData)
        {
          Serial.printf("/open file ok '%s'\n", fileName);
 
          
          // close file in callback readDataJsonFromSD
          AsyncWebServerResponse *response = request->beginResponse(String("text/plain"), g_fileTempData.size(), readTempDataFromSD);
          response->addHeader("Server", "MyServerString");
          request->send(response);
          return;
        }
        // takeSaveFilesAccess was succefull, but SD.open not --> release Semaphore
        handleSD::GetInstance().releaseSaveFileAccess();
      }
      xSemaphoreGive(m_SemaphoreFile); // can not open file, release semaphore, return 404
    }
    request->send(404); // if no semaphore return 404
  }); 

  ////////////////////////////////////////////////////////////////////////
  server.on("/list" , HTTP_GET, [](AsyncWebServerRequest * request)
  {
    Serial.printf("/list\n");

    int paramsNr = request->params();
    String rootDir = "/";
    int level = 1; // 1 subdirectory
   
    for(int i=0;i<paramsNr;i++)
    {
      AsyncWebParameter* p = request->getParam(i);
      if(p->name() == "root")
      {
        rootDir = p->value();
      }
      else if(p->name() == "level")
      {
        level = p->value().toInt();
      }
    }    
    handleSD& instanceSD = handleSD::GetInstance();
    String stringReturn = instanceSD.listDir(&rootDir[0], level);
    stringReturn += String("\nSyntax: http://<server>/list[?root=/2019/][&level=2] --> print folder /2019/ and two subfolder recursivly\n");
    
    request->send(200, "text/plain", &stringReturn[0]);
  });

#ifdef LOG_BOOT_TIME
  ////////////////////////////////////////////////////////////////////////
  server.on("/system" , HTTP_GET, [](AsyncWebServerRequest * request)
  {
    // protect from other web request's
    if(xSemaphoreTake(m_SemaphoreFile, 2000/portTICK_PERIOD_MS ) == pdTRUE ) // wait max 2 seconds
    {
        WebDebug::GetInstance().system_info();
        snprintf(WebDebug::GetInstance().getFreeBuffer(), DEBUG_LINE_LENGTH, "SD Time: %d", handleSD::GetInstance().getSDTime());
        
        // close file in callback printDebugData
        AsyncWebServerResponse *response = request->beginResponse(String("text/plain"),WebDebug::GetInstance().CreateSortedRingbuffer(), WebDebug::printDebugData);
        response->addHeader("Server", "MyServerString");
        request->send(response);
        return;
    }
    request->send(404); // if no semaphore return 404
  });
  
  ////////////////////////////////////////////////////////////////////////
  server.on("/boot" , HTTP_GET, [](AsyncWebServerRequest * request)
  {
    display1306& display = display1306::GetInstance();
    display.addMessage(String("Boot called\n"), PRIO_DEBUG); // display message on oled for 5 seconds
    
    // protect from other web request's
    if(xSemaphoreTake(m_SemaphoreFile, 2000/portTICK_PERIOD_MS ) == pdTRUE ) // wait max 2 seconds
    {
      char fileName[21];
      snprintf(fileName, sizeof(fileName), "/boot.txt");
    
      // check if file is now open for writing
      if(handleSD::GetInstance().takeSaveFilesAccess(100/portTICK_PERIOD_MS)) // wait max 100 ms for file access
      {
        g_fileTempData = SD.open(fileName, FILE_READ);
        if(g_fileTempData)
        {
          // close file in callback readDataJsonFromSD
          AsyncWebServerResponse *response = request->beginResponse(String("text/plain"), g_fileTempData.size(), readTempDataFromSD);
          response->addHeader("Server", "MyServerString");
          request->send(response);
          return;
        }
        // takeSaveFilesAccess was succefull, but SD.open not --> release Semaphore
        handleSD::GetInstance().releaseSaveFileAccess();
      }
      xSemaphoreGive(m_SemaphoreFile); // can not open file, release semaphore, return 404
    }
    request->send(404); // if no semaphore return 404
  });

#endif

 /* 
  ////////////////////////////////////////////////////////////////////////
  server.on("/delete" , HTTP_GET, [](AsyncWebServerRequest * request)
  {
    String fileName = "Can not delete ... ";
    
    DateTime dt(MyNTPClient::GetInstance().getEpochTime());

    for(int i = 0; i < request->params(); i++)
    {
      AsyncWebParameter* p = request->getParam(i);
      Serial.print("Parameter Name: ");
      Serial.println(p->name());
      Serial.print("Parameter Value: ");
      Serial.println(p->value());
      Serial.println("------");

      if(p->name() == "file")
      {
        fileName = p->value();
        if(handleSD::GetInstance().deleteFile(&fileName[0]))
        {
          fileName = String("Delete ") + fileName + String(" OK.");
        }
        else
        {
          fileName = String("Delete ") + fileName + String(" failed.");
        }
      }
    }
    request->send(200, "text/plain", fileName);
  });*/
    
  return 0;
}
