/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.
****************************************************************************************/ 
/*
 * Connect the SD card to the following pins:
 *
// SD - GPIO12 - SCL
// SD - GPIO15 - MISO 
// SD - GPIO13 - MOSI
// SD - GPIO14 - CS
 */

#include "handleSD.h"
#include "display1306.h"
#include "sensor18b20.h"
#include "tempLoggerESP32.h"
#include "myNTPClient.h"
 
//////////////////////////////////////////////////////////////////////
handleSD& handleSD::GetInstance()
{
  static handleSD instance;
  return instance;
}

//////////////////////////////////////////////////////////////////////
handleSD::handleSD()
{
  m_xRWSemaphore = xSemaphoreCreateMutex();
  m_fileTimes[0] = millis();
  m_fileTimes[1] = millis();
  m_lastDay = -1;
}

//////////////////////////////////////////////////////////////////////
void handleSD::setup()
{
  //SPI.begin(18,19,23); if(!SD.begin(5)
  //  void begin(int8_t sck=-1, int8_t miso=-1, int8_t mosi=-1, int8_t ss=-1);
  SPI.begin(12, 15, 13);   
  
  // try 10 times to initialize
  int i = 0;
  for(i = 0; i < 10; i++)
  {
    if(SD.begin(14, SPI)) {
      break;
    }
    Serial.printf("Card Mount Failed %d\n", i);
    display1306::GetInstance().addMessage(String("Card Mount Failed ") + String(i), PRIO_FATAL);
  
    delay(150);
  }
  if (i==10)
  {
    return;
  }
 
  if(CARD_NONE == SD.cardType())
  {
      Serial.println("No SD card attached");
      display1306::GetInstance().addMessage(String("No SD card attached"), PRIO_FATAL);
  
      return;
  }

  m_NextLoop = (int(MyNTPClient::GetInstance().getEpochTime() / 60)+1)*60; // round to next minute  
}

//////////////////////////////////////////////////////////////////////
void handleSD::writeToSD()
{
  unsigned long nowTime = MyNTPClient::GetInstance().getEpochTime();
  DateTime dt(nowTime);
  
  // somtimes the time is 2036/02/07 ... integrate other NTPClient library - bug should not longer accurs
  if(dt.year() == 2036)
  {
    return;
  }

  // reset time if diff more than 60 seconds, 
  // somtimes the time is 2036/02/07 ... bug should not longer accurs
  if(abs(m_NextLoop - nowTime) > 100)
  {
    m_NextLoop = ((nowTime / 60)+1)*60; // round to next minute  
  }

  if(nowTime >= m_NextLoop)
  {    
    m_NextLoop = ((nowTime / 60)+1)*60; // round to next minute  
    Serial.print(MyNTPClient::GetInstance().getFormattedTime());
    //Serial.println(" #### SD write ####");

    Serial.printf("handleSD::writeToSD xSemaphoreTake\n");

    sensor18b20& instanceSensor = sensor18b20::GetInstance();
    DateTime dt(MyNTPClient::GetInstance().getEpochTime());

    // for all sensors
    for ( byte thisSensor = 0; thisSensor < instanceSensor.getNumberSensorCnt(); thisSensor++)
    {
      // unchanged text and the same day --> next sensor in loop
      if(!instanceSensor.isSensorEntryTextChanged(thisSensor) && (dt.day() == m_lastDay))
      {
   //     Serial.printf("Ignore same value %s: %s\n", instanceSensor.getSensorCRC16Id(thisSensor).c_str(), instanceSensor.getSensorEntryText(thisSensor).c_str());
        continue;
      }

     // Serial.printf("Write value %s: %s\n", instanceSensor.getSensorCRC16Id(thisSensor).c_str(), instanceSensor.getSensorEntryText(thisSensor).c_str());
     
      String fileName = buildPath(dt.year(), dt.month(), dt.day(), instanceSensor.getSensorCRC16Id(thisSensor));
      if(fileName.length() > 0)
      {
        String data = instanceSensor.getSensorEntryText(thisSensor);
        Serial.printf("Write file %s --> %s\n", &fileName[0], &data[0]);
 
        if(takeSaveFilesAccess(5000/portTICK_PERIOD_MS)) // wait max 5 seconds for writing file
        {
          m_fileTimes[0] = millis();
          File file = SD.open(fileName, FILE_APPEND);
          if(!file)
          {
              Serial.println("Failed to open file for appending");
              display1306::GetInstance().addMessage(String("Failed to open file for appending ")+String(fileName), PRIO_FATAL);
 
              m_fileTimes[1] = millis();
              return;
          }

          // schreibe "Stunden:Minuten;Temperatur*100"
          char content[14];
          snprintf(content, sizeof(content), "%02d:%02d;%s", dt.hour(), dt.minute(), &instanceSensor.getSensorEntryText(thisSensor)[0]);
     
          if(!file.println(String(content)))
          {
              Serial.println("Append failed");
          }
          file.close();
          m_fileTimes[1] = millis();
          releaseSaveFileAccess();
        }
      }
    }
    // set actual day for new day check 
    m_lastDay = dt.day();

  }
#ifdef LOG_BOOT_TIME
  static bool bWriteStartUpTime = false;
  if(!bWriteStartUpTime && MyNTPClient::GetInstance().isTimeSynchronized())
  {
    bWriteStartUpTime = true;  // write boot time into file
    writeBootTime();
  }
#endif    
}

//////////////////////////////////////////////////////////////////////
boolean handleSD::takeSaveFilesAccess(TickType_t xTicksToWait)
{
  return (xSemaphoreTake(m_xRWSemaphore, xTicksToWait) == pdTRUE );
}

//////////////////////////////////////////////////////////////////////
void handleSD::releaseSaveFileAccess()
{
  xSemaphoreGive(m_xRWSemaphore); 
}

//////////////////////////////////////////////////////////////////////
String handleSD::buildPath(int year, int month, int day, String sensorId)
{
  char cYear[5];
  char cMonth[3];
  char fileName[12];
  snprintf(cYear, sizeof(cYear), "%d", year);
  snprintf(cMonth, sizeof(cMonth), "%02d", month);
  snprintf(fileName, sizeof(fileName), "%02d_%s.txt", day, &sensorId[0]);

  Serial.printf("%s/%s/%s\n", cYear, cMonth, fileName);

  String path = String("/") + String(cYear);
  //Serial.printf("mkdir %s\n", &path[0]);
  if(SD.mkdir(&path[0]))
  {
    path += String("/") + String(cMonth);
    //Serial.printf("mkdir %s\n", &path[0]);
    if(SD.mkdir(&path[0]))
    {
      // 7.3
      path += String("/") + String(fileName);
      return path;
    }
  }

  return String("");
}

//////////////////////////////////////////////////////////////////////
bool handleSD::deleteFile(char* path)
{
  bool bRc = true;
  
  Serial.printf("Deleting file: %s\n", path);
  
  if(SD.remove(path))
  {
    Serial.println("File deleted");
  } 
  else 
  {
    if(SD.rmdir(path))
    {
      Serial.println("Directory deleted");
    } 
    else
    {
      Serial.println("Delete failed");
      bRc = false;      
    }
  }
  return bRc;
}

//////////////////////////////////////////////////////////////////////
String handleSD::listDir(const char * dirname, uint8_t levels)
{
    String rcString = "";
    
    Serial.printf("Listing directory: %s\n", dirname);
  
    File root = SD.open(dirname);
    m_fileTimes[0] = millis();
    if(!root)
    {
        Serial.println("Failed to open directory");
        return rcString;
    }
    if(!root.isDirectory()){
        Serial.println("Not a directory");
        return rcString;
    }

    File file = root.openNextFile();
    while(file)
    {
        if(file.isDirectory())
        {
            Serial.print("  DIR : ");
            Serial.println(file.name());
            rcString += String("  DIR : ");
            rcString += file.name() + String("\n");            
            if(levels)
            {
                rcString += listDir(file.name(), levels -1);
            }
        } 
        else 
        {
            Serial.print("  FILE: ");
            Serial.print(file.name());
            Serial.print("  SIZE: ");
            Serial.println(file.size());
            rcString += String("  FILE: ");
            rcString += String(file.name());
            rcString += String("  SIZE: ");
            rcString += file.size() + String("\n");
        }
        file = root.openNextFile();
    }
    return rcString;
}

//////////////////////////////////////////////////////////////////////
unsigned long handleSD::getSDTime()
{
  return (m_fileTimes[1] - m_fileTimes[0]);
}

//////////////////////////////////////////////////////////////////////
void handleSD::writeBootTime()
{
  MyNTPClient& ntpClient = MyNTPClient::GetInstance();
  
  unsigned long nowTime = ntpClient.getEpochTime();
  DateTime dt(nowTime);

  if(takeSaveFilesAccess(5000/portTICK_PERIOD_MS)) // wait max 5 seconds for writing file
  {
    char *name1 = "/boot.txt";
    char *name2 = "/boot.tmp";
    uint8_t cpBuffer[256];
    int maxLen = 2*1024;
    
    SD.remove(name2);
    if(!SD.rename(name1,name2))
    {
      Serial.printf("writeBootTime rename failed %s, %s\n", name1, name2);
    }
    
    File fileRead = SD.open(name2);
    if(!fileRead)
    {
      Serial.printf("writeBootTime rename failed %s\n", name2);
    }

    size_t fileSize = fileRead.size();
    File fileWrite = SD.open(name1, FILE_APPEND);
    if(!fileWrite)
    {
      Serial.printf("writeBootTime open append failed %s\n", name1);
    }

    // ignore start of file
    while(fileRead.available() && (--fileSize > maxLen))
    {
      fileRead.read();
    }

    // copy part of file
    while(fileRead.available())
    {
        if(fileSize > sizeof(cpBuffer))
        {
          fileWrite.write(cpBuffer, fileRead.read(cpBuffer, sizeof(cpBuffer)));
        }
        else
        {
          fileWrite.write(cpBuffer, fileRead.read(cpBuffer, fileSize));
          break;
        }
        fileSize -= sizeof(cpBuffer);
    }
    fileRead.close();

    // append new entry
    char content[30];
    snprintf(content, sizeof(content), "%04d/%02d/%02d %02d:%02d:%02d  - ", dt.year(), dt.month(), dt.day(), dt.hour(), dt.minute(), dt.second());
    if(!fileWrite.print(String(content)))
    {
        Serial.println("Append failed");
    }
     
    //https://github.com/espressif/arduino-esp32/blob/master/libraries/ESP32/examples/ResetReason/ResetReason.ino
    for(int i=0;i<2;i++)
    {
      char *reason="";
      switch(rtc_get_reset_reason(i))
      {
        case 1 : reason = "POWERON_RESET";break;          /**<1, Vbat power on reset*/
        case 3 : reason = "SW_RESET";break;               /**<3, Software reset digital core*/
        case 4 : reason = "OWDT_RESET";break;             /**<4, Legacy watch dog reset digital core*/
        case 5 : reason = "DEEPSLEEP_RESET";break;        /**<5, Deep Sleep reset digital core*/
        case 6 : reason = "SDIO_RESET";break;             /**<6, Reset by SLC module, reset digital core*/
        case 7 : reason = "TG0WDT_SYS_RESET";break;       /**<7, Timer Group0 Watch dog reset digital core*/
        case 8 : reason = "TG1WDT_SYS_RESET";break;       /**<8, Timer Group1 Watch dog reset digital core*/
        case 9 : reason = "RTCWDT_SYS_RESET";break;       /**<9, RTC Watch dog Reset digital core*/
        case 10 : reason = "INTRUSION_RESET";break;       /**<10, Instrusion tested to reset CPU*/
        case 11 : reason = "TGWDT_CPU_RESET";break;       /**<11, Time Group reset CPU*/
        case 12 : reason = "SW_CPU_RESET";break;          /**<12, Software reset CPU*/
        case 13 : reason = "RTCWDT_CPU_RESET";break;      /**<13, RTC Watch dog Reset CPU*/
        case 14 : reason = "EXT_CPU_RESET";break;         /**<14, for APP CPU, reseted by PRO CPU*/
        case 15 : reason = "RTCWDT_BROWN_OUT_RESET";break;/**<15, Reset when the vdd voltage is not stable*/
        case 16 : reason = "RTCWDT_RTC_RESET";break;      /**<16, RTC Watch dog reset digital core and rtc module*/
        default : reason = "NO_MEAN";
      }
      snprintf(content, sizeof(content),"%d: %s ", i, reason);
      fileWrite.print(String(content));
    }
    fileWrite.println(String(""));

    fileWrite.close();
    releaseSaveFileAccess();
  }
}
