/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.

  Based on:
  // OneWire DS18S20, DS18B20, DS1822 Temperature Example
  //
  // http://www.pjrc.com/teensy/td_libs_OneWire.html
  //
  // The DallasTemperature library can do all this work for you!
  // http://milesburton.com/Dallas_Temperature_Control_Library
****************************************************************************************/


#include "OneWire.h"
#include "Crc16.h"

//#define SHOW_DALLAS_ERROR        // uncomment to show Dallas ( CRC ) errors on Serial.
#define ONEWIRE_PIN           25   // OneWire Dallas sensors are connected to this pin
#define MAX_NUMBER_OF_SENSORS 50   // maximum number of Dallas sensors

OneWire  ds( ONEWIRE_PIN );        // (a 4.7K pull-up resistor is necessary)


byte m_MeasureIndex = 0;
byte m_ProcessIndex = 0;
byte m_numberOfFoundSensors = 0;
char m_TextSensor1[50];
char m_TextSensor2[50];
extern char *pSensor1;
extern char *pSensor2;

void loop_18b20();

//////////////////////////////////////////////////////////////////////
struct SensorData
{
  byte    m_Addr[8];
  String  m_crc16Name;
  float   m_tempCelcius;
};

// Model IDs
#define DS18S20MODEL 0x10  // also DS1820
#define DS18B20MODEL 0x28
#define DS1822MODEL  0x22
#define DS1825MODEL  0x3B

#define MAX_NUMBER_OF_SENSORS 20   // maximum number of Dallas sensors
SensorData  m_Sensor[MAX_NUMBER_OF_SENSORS];


//////////////////////////////////////////////////////////////////////
void run_18b20(void * pvParameters)
{
  byte currentAddr[8];

  ds.reset_search();
  while ( ds.search( currentAddr ) && m_numberOfFoundSensors < MAX_NUMBER_OF_SENSORS )
  {
    // set m_Addr
    memcpy(m_Sensor[m_numberOfFoundSensors].m_Addr, currentAddr, sizeof(currentAddr));
    m_numberOfFoundSensors++;
  }

  Serial.printf( "%i Dallas sensors found.\n", m_numberOfFoundSensors );

  if ( !m_numberOfFoundSensors )
  {
    pSensor1 = "No Sensor found";
    vTaskDelete( NULL );
  }
  else
  {
    // set resolution to 12 bit
    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      /////////////
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0x4E );  // Write Scratchpad
      ds.write( 0x4b );  // high alarm temp, dummy value
      ds.write( 0x46 );  // low alarm temp, dummy value
      // DS1820 and DS18S20 have no configuration register
      if (m_Sensor[thisSensor].m_Addr[0] != DS18S20MODEL) 
      {
         ds.write( 0x7F );  // 12 bit
      }
      
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0x48, false );  // COPYSCRATCH, parasite
      delay(20);  // <--- added 20ms delay to allow 10ms long EEPROM write operation (as specified by datasheet)
      ds.reset();
    }

    Crc16 crc;            // build CRC16  
    char crc16Buffer[5];  // build CRC16  
    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      snprintf( crc16Buffer, sizeof(crc16Buffer), "%04X", crc.XModemCrc(m_Sensor[thisSensor].m_Addr, 0, 8));    
      m_Sensor[thisSensor].m_crc16Name = String(crc16Buffer);      
    }

    
  }
  loop_18b20();
}

//////////////////////////////////////////////////////////////////////
void loop_18b20()
{
  pSensor1 = &m_TextSensor1[0];
  pSensor2 = &m_TextSensor2[0];

  /* main temptask loop */
  while (1)
  {
    vTaskDelay( 7000 / portTICK_PERIOD_MS);
    
    // trigger measurement
    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0x44, 0);        // start conversion, with parasite power off at the end
    }

    //wait for conversion ready
    vTaskDelay( 3*1000 / portTICK_PERIOD_MS); 

    for ( byte thisSensor = 0; thisSensor < m_numberOfFoundSensors; thisSensor++)
    {
      byte crcRetry = 0;
      byte data[12];
      ds.reset();
      ds.select( m_Sensor[thisSensor].m_Addr );
      ds.write( 0xBE );         // Read Scratchpad

      //Serial.print( "Sensor " );Serial.print( thisSensor ); Serial.print("  Data = ");
      //Serial.println( present, HEX );
      //Serial.print(" ");
      for ( byte i = 0; i < 9; i++)
      { // we need 9 bytes
        data[i] = ds.read();
        //Serial.print(data[i], HEX);
        //Serial.print(" ");
      }
      //Serial.println();

      byte type_s;
      // the first ROM byte indicates which chip
      switch ( m_Sensor[thisSensor].m_Addr[0] )
      {
        case 0x10:
          //Serial.println("  Chip = DS18S20");  // or old DS1820
          type_s = 1;
          break;
        case 0x28:
          //Serial.println("  Chip = DS18B20");
          type_s = 0;
          break;
        case 0x22:
          //Serial.println("  Chip = DS1822");
          type_s = 0;
          break;
        default:
#ifdef SHOW_DALLAS_ERROR
          Serial.println("Device is not a DS18x20 family device.");
#endif
          return;
      }

      int16_t raw;
      if ( OneWire::crc8(data, 8) != data[8])
      {
#ifdef SHOW_DALLAS_ERROR
        // CRC of temperature reading indicates an error, so we print a error message and discard this reading
        Serial.print( millis() / 1000.0 ); Serial.print( " - CRC error from device " ); Serial.println( thisSensor );
#endif
        if(crcRetry < 2)
        {
          crcRetry++;
          thisSensor--;          
        }
      }
      else
      {
        raw = (data[1] << 8) | data[0];
        if (type_s)
        {
          raw = raw << 3; // 9 bit resolution default
          if (data[7] == 0x10)
          {
            // "count remain" gives full 12 bit resolution
            raw = (raw & 0xFFF0) + 12 - data[6];
          }
        }
        else
        {
          byte cfg = (data[4] & 0x60);
          // at lower res, the low bits are undefined, so let's zero them
          if (cfg == 0x00) raw = raw & ~7;  // 9 bit resolution, 93.75 ms
          else if (cfg == 0x20) raw = raw & ~3; // 10 bit res, 187.5 ms
          else if (cfg == 0x40) raw = raw & ~1; // 11 bit res, 375 ms
          //// default is 12 bit resolution, 750 ms conversion time
        }
        m_Sensor[thisSensor].m_tempCelcius = raw / 16.0;
      }
    }
    
    Serial.printf("measure %.2f\n", m_Sensor[0].m_tempCelcius);
    sprintf(m_TextSensor1, "%d Sensors found", m_numberOfFoundSensors);
    sprintf(m_TextSensor2, "Id: %s = %.2f", m_Sensor[0].m_crc16Name.c_str(), m_Sensor[0].m_tempCelcius);
    
    m_MeasureIndex++; // measure loop finished

    // loop 255  
    if(m_MeasureIndex == (byte)-1)
    {
      m_MeasureIndex = 1;
      m_ProcessIndex = 0;    
    }
  } // end while
}




